@@ -1,4 +1,4 @@
-# Copyright (c) 2020, Riverbank Computing Limited
+# Copyright (c) 2021, Riverbank Computing Limited
 # All rights reserved.
 #
 # This copy of SIP is licensed for use under the terms of the SIP License
@@ -21,13 +21,13 @@
 # POSSIBILITY OF SUCH DAMAGE.
 
 
-from collections import OrderedDict
-from distutils.sysconfig import get_python_inc, get_python_lib
+import collections
 import os
 import packaging
 import shutil
 import subprocess
 import sys
+import sysconfig
 import tempfile
 import warnings
 
@@ -37,7 +37,7 @@ from .bindings import Bindings
 from .configurable import Configurable, Option
 from .exceptions import UserException
 from .module import resolve_abi_version
-from .py_versions import FIRST_SUPPORTED_MINOR, LAST_SUPPORTED_MINOR
+from .py_versions import OLDEST_SUPPORTED_MINOR
 from .pyproject import (PyProjectException, PyProjectOptionException,
         PyProjectUndefinedOptionException)
 
@@ -63,15 +63,22 @@ class Project(AbstractProject, Configurable):
         # Set if an __init__.py should be installed.
         Option('dunder_init', option_type=bool, default=False),
 
+        # The list of GUI script entry points.
+        Option('gui_scripts', option_type=list),
+
         # The minimum GLIBC version required by the project.  This is used to
         # determine the correct platform tag to use for Linux wheels.
         Option('minimum_glibc_version'),
 
+        # The minimum macOS version required by the project.  This is used to
+        # determine the correct platform tag to use for macOS wheels.
+        Option('minimum_macos_version'),
+
         # Set if building for a debug version of Python.
         Option('py_debug', option_type=bool),
 
         # The name of the directory containing Python.h.
-        Option('py_include_dir', default=get_python_inc()),
+        Option('py_include_dir', default=sysconfig.get_path('include')),
 
         # The name of the target Python platform.
         Option('py_platform'),
@@ -117,11 +124,16 @@ class Project(AbstractProject, Configurable):
         Option('build_dir', help="the build directory", metavar="DIR"),
         Option('build_tag', help="the build tag to be used in the wheel name",
                 metavar="TAG", tools=['wheel']),
+        Option('distinfo', option_type=bool, inverted=True,
+                help="don't create a .dist-info directory", tools=['install']),
         Option('manylinux', option_type=bool, inverted=True,
                 help="disable the use of manylinux in the platform tag used "
                         "in the wheel name",
                 tools=['wheel']),
-        Option('target_dir', default=get_python_lib(plat_specific=1),
+        Option('scripts_dir', default=os.path.dirname(sys.executable),
+                help="the scripts installation directory", metavar="DIR",
+                tools=['build', 'install']),
+        Option('target_dir', default=sysconfig.get_path('platlib'),
                 help="the target installation directory", metavar="DIR",
                 tools=['build', 'install']),
         Option('api_dir', help="generate a QScintilla .api file in DIR",
@@ -139,18 +151,21 @@ class Project(AbstractProject, Configurable):
     def __init__(self, **kwargs):
         """ Initialise the project. """
 
-        super().__init__(**kwargs)
+        super().__init__()
 
         # The current directory should contain the .toml file.
         self.root_dir = os.getcwd()
-        self.bindings = OrderedDict()
+        self.bindings = collections.OrderedDict()
         self.bindings_factories = []
         self.builder = None
         self.buildables = []
         self.installables = []
 
+        self._metadata_overrides = None
         self._temp_build_dir = None
 
+        self.initialise_options(kwargs)
+
     def apply_nonuser_defaults(self, tool):
         """ Set default values for non-user options that haven't been set yet.
         """
@@ -162,18 +177,22 @@ class Project(AbstractProject, Configurable):
             self.bindings_factory = self.import_callable(self.bindings_factory,
                     Bindings)
 
+        if self.py_major_version is None or self.py_minor_version is None:
+            self.py_major_version = sys.hexversion >> 24
+            self.py_minor_version = (sys.hexversion >> 16) & 0x0ff
+
         if self.builder_factory is None:
-            from .distutils_builder import DistutilsBuilder
-            self.builder_factory = DistutilsBuilder
+            if (self.py_major_version, self.py_minor_version) >= (3, 10):
+                from .setuptools_builder import SetuptoolsBuilder
+                self.builder_factory = SetuptoolsBuilder
+            else:
+                from .distutils_builder import DistutilsBuilder
+                self.builder_factory = DistutilsBuilder
         elif isinstance(self.builder_factory, str):
             # Convert the name to a callable.
             self.builder_factory = self.import_callable(self.builder_factory,
                     AbstractBuilder)
 
-        if self.py_major_version is None or self.py_minor_version is None:
-            self.py_major_version = sys.hexversion >> 24
-            self.py_minor_version = (sys.hexversion >> 16) & 0x0ff
-
         if self.py_platform is None:
             self.py_platform = sys.platform
 
@@ -185,8 +204,8 @@ class Project(AbstractProject, Configurable):
     def apply_user_defaults(self, tool):
         """ Set default values for user options that haven't been set yet. """
 
-        # If we the backend to a 3rd-party frontend (most probably pip) then
-        # let it handle the verbosity of messages.
+        # If we are the backend to a 3rd-party frontend (most probably pip)
+        # then let it handle the verbosity of messages.
         if self.verbose is None and tool == '':
             self.verbose = True
 
@@ -263,6 +282,15 @@ class Project(AbstractProject, Configurable):
         # This default implementation will create an empty file.
         return ''
 
+    def get_metadata_overrides(self):
+        """ Return a mapping of PEP 566 metadata names and values that will
+        override any corresponding values defined in the pyproject.toml file.
+        A typical use is to determine a project's version dynamically.
+        """
+
+        # This default implementation does not override any metadata.
+        return {}
+
     def get_options(self):
         """ Return the list of configurable options. """
 
@@ -312,6 +340,56 @@ class Project(AbstractProject, Configurable):
         return ['{} (>={}, <{})'.format(sip_project_name, self.abi_version,
                 next_abi_major)]
 
+    def get_sip_distinfo_command_line(self, sip_distinfo, inventory,
+            generator=None, wheel_tag=None):
+        """ Return a sequence of command line arguments to invoke sip-distinfo.
+        """
+
+        args = [
+            sip_distinfo,
+
+            '--inventory',
+            inventory,
+
+            '--project-root',
+            self.root_dir,
+
+            '--prefix',
+            '\\"$(INSTALL_ROOT)\\"',
+        ]
+
+        if generator is not None:
+            args.append('--generator')
+            args.append(generator)
+
+        if wheel_tag is not None:
+            args.append('--wheel-tag')
+            args.append(wheel_tag)
+
+        for ep in self.console_scripts:
+            args.append('--console-script')
+            args.append(ep.replace(' ', ''))
+
+        for ep in self.gui_scripts:
+            args.append('--gui-script')
+            args.append(ep.replace(' ', ''))
+
+        for rd in self.get_requires_dists():
+            args.append('--requires-dist')
+            args.append('\\"{}\\"'.format(rd))
+
+        for metadata, value in self._metadata_overrides.items():
+            if value:
+                metadata += '=' + value
+
+            if ' ' in metadata:
+                metadata = '\\"' + metadata + '\\"'
+
+            args.append('--metadata')
+            args.append(metadata)
+
+        return args
+
     def install(self):
         """ Install the project. """
 
@@ -459,6 +537,9 @@ class Project(AbstractProject, Configurable):
             if not b.is_buildable():
                 del self.bindings[b.name]
 
+        if len(self.bindings) == 0:
+            raise UserException("There are no bindings that can be built")
+
     def verify_configuration(self, tool):
         """ Verify that the configuration is complete and consistent. """
 
@@ -471,13 +552,9 @@ class Project(AbstractProject, Configurable):
         # Make sure any minimum GLIBC version is valid and convert it to a
         # 2-tuple.
         if self.minimum_glibc_version:
-            parts = self.minimum_glibc_version.split('.')
-
             try:
-                if len(parts) != 2:
-                    raise ValueError()
-
-                self.minimum_glibc_version = (int(parts[0]), int(parts[1]))
+                self.minimum_glibc_version = self._convert_major_minor(
+                        self.minimum_glibc_version)
             except ValueError:
                 raise PyProjectOptionException('minimum-glibc-version',
                         "'{0}' is an invalid GLIBC version number".format(
@@ -486,26 +563,40 @@ class Project(AbstractProject, Configurable):
         else:
             self.minimum_glibc_version = (2, 5)
 
+        # Make sure any minimum macOS version is valid and convert it to a
+        # 2-tuple.
+        if self.minimum_macos_version:
+            try:
+                self.minimum_macos_version = self._convert_major_minor(
+                        self.minimum_macos_version)
+            except ValueError:
+                raise PyProjectOptionException('minimum-macos-version',
+                        "'{0}' is an invalid macOS version number".format(
+                                self.minimum_macos_version),
+                        section_name='tool.sip.project')
+
         # Make sure relevent paths are absolute and use native separators.
         self.sip_files_dir = self.project_path(self.sip_files_dir)
         self.sip_include_dirs = [self.project_path(d)
                 for d in self.sip_include_dirs]
 
-        # Make sure we support the targeted version of Python.
+        # Check that the targeted version of Python isn't too old.  We hope
+        # that we will support newer versions automatically, but it's not
+        # guaranteed.
         py_version = (self.py_major_version, self.py_minor_version)
-        first_version = (3, FIRST_SUPPORTED_MINOR)
-        last_version = (3, LAST_SUPPORTED_MINOR)
+        first_version = (3, OLDEST_SUPPORTED_MINOR)
 
-        if py_version < first_version or py_version > last_version:
+        if py_version < first_version or self.py_major_version > 3:
             raise UserException(
                     "Python v{}.{} is not supported".format(
                             self.py_major_version, self.py_minor_version))
 
-        # Make sure we have a valid ABI version.
-        self.abi_version = resolve_abi_version(self.abi_version)
+        # Get the supported ABI version.  (The actual version may have a later
+        # minor version.)
+        self.abi_version = resolve_abi_version(self.abi_version, module=False)
 
         # Checks for standalone projects.
-        if not self.sip_module:
+        if tool in Option.BUILD_TOOLS and not self.sip_module:
             # Check there is only one set of bindings.
             if len(self.bindings) > 1:
                 raise PyProjectOptionException('sip-module',
@@ -583,6 +674,18 @@ class Project(AbstractProject, Configurable):
                     setattr(configurable, option.name,
                             getattr(args, option.dest))
 
+    @staticmethod
+    def _convert_major_minor(value):
+        """ Convert a 'major.minor' version number to a 2-tuple of integers.
+        Raise a ValueError exception if it is invalid.
+        """
+
+        parts = value.split('.')
+        if len(parts) != 2:
+            raise ValueError()
+
+        return int(parts[0]), int(parts[1])
+
     def _enable_disable_bindings(self):
         """ Check the enabled bindings are valid and remove any disabled ones.
         """
@@ -594,7 +697,7 @@ class Project(AbstractProject, Configurable):
             for enabled in self.enable:
                 if enabled not in names:
                     raise UserException(
-                            "unknown enabled bindings '{0}'".format(enabled))
+                            "Unknown enabled bindings '{0}'".format(enabled))
 
             # Only include explicitly enabled bindings.
             for b in list(self.bindings.values()):
@@ -606,7 +709,7 @@ class Project(AbstractProject, Configurable):
             for disabled in self.disable:
                 if disabled not in names:
                     raise UserException(
-                            "unknown disabled bindings '{0}'".format(disabled))
+                            "Unknown disabled bindings '{0}'".format(disabled))
 
             # Remove any explicitly disabled bindings.
             for b in list(self.bindings.values()):
@@ -623,6 +726,8 @@ class Project(AbstractProject, Configurable):
 
         # Get the metadata and extract the version.
         self.metadata = pyproject.get_metadata()
+        self._metadata_overrides = self.get_metadata_overrides()
+        self.metadata.update(self._metadata_overrides)
         self.version_str = self.metadata['version']
 
         # Convert the version as a string to number.
